/*
 *	plugin_monkey.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spInputPluginP.h>

#include <MAC/All.h>
#include <MAC/MACLib.h>
#include <MAC/APETag.h>
#if 1
#ifdef _WIN32
#include <MAC/WinFileIO.h>
#ifndef IO_CLASS_NAME
#define IO_CLASS_NAME CWinFileIO
#endif
#else
#include <MAC/StdLibFileIO.h>
#ifndef IO_CLASS_NAME
#define IO_CLASS_NAME CStdLibFileIO
#endif
#endif
#else
#include <MAC/IO.h>
#endif

#if !defined(APE_VERSION_NUMBER) && defined(MAC_VERSION_NUMBER)
#define APE_VERSION_NUMBER MAC_VERSION_NUMBER
#endif
#if !defined(APE_FILE_VERSION_NUMBER) && defined(MAC_FILE_VERSION_NUMBER)
#define APE_FILE_VERSION_NUMBER MAC_FILE_VERSION_NUMBER
#endif

#if !defined(APE_COMPRESSION_LEVEL_HIGH)
#if defined(MAC_COMPRESSION_LEVEL_HIGH)
#define APE_COMPRESSION_LEVEL_HIGH MAC_COMPRESSION_LEVEL_HIGH
#else
#define APE_COMPRESSION_LEVEL_HIGH COMPRESSION_LEVEL_HIGH
#endif
#endif

#if APE_FILE_VERSION_NUMBER >= 3980
using namespace APE;

#ifdef _WIN32
#include <mbstring.h>
#else
#include <wchar.h>
#define _mbslen(s) mblen((const char *)(s), MB_CUR_MAX)
#define _wcsicmp wcscasecmp
#endif
/*typedef str_utf16 APEChar;*/
typedef str_utfn APEChar;
#else
typedef char APEChar;
#endif

#define MONKEY_DEFAULT_BUFFER_SIZE 2048

#if APE_FILE_VERSION_NUMBER >= 3990
#define APEBool bool 
#define APE_TRUE true
#define APE_FALSE false
#else
#define APEBool BOOL
#define APE_TRUE TRUE
#define APE_FALSE FALSE
#endif

#if APE_VERSION_MAJOR >= 8
#define APE_INFO_TAG IAPEDecompress::APE_INFO_TAG
#define APE_INFO_COMPRESSION_LEVEL IAPEDecompress::APE_INFO_COMPRESSION_LEVEL
#define APE_INFO_CHANNELS IAPEDecompress::APE_INFO_CHANNELS
#define APE_INFO_SAMPLE_RATE IAPEDecompress::APE_INFO_SAMPLE_RATE
#define APE_INFO_BITS_PER_SAMPLE IAPEDecompress::APE_INFO_BITS_PER_SAMPLE
#define APE_INFO_FORMAT_FLAGS IAPEDecompress::APE_INFO_FORMAT_FLAGS
#define APE_DECOMPRESS_TOTAL_BLOCKS IAPEDecompress::APE_DECOMPRESS_TOTAL_BLOCKS
#endif

typedef struct _spMonkeyPluginInstance
{
    IO_CLASS_NAME FileIO;
    IAPEDecompress * pAPEDecompress;
    IAPECompress * pAPECompress;

    int compression_level;

    int num_channel;
    long samp_rate;
    long samp_bit;

    char filename[SP_MAX_PATHNAME];
    spLong current_pos;
    spLong total_length;

    spSongInfo song_info;

    unsigned char *buffer;
    spLong buffer_size;
    spLong buffer_length;
    int buffer_samp_byte;
} *spMonkeyPluginInstance;

static spBool spInitPluginMonkey(const char *lang)
{
    /*spSetDebugLevel(100);*/
    return SP_TRUE;
}

static spBool spFreePluginMonkey(void)
{
    return SP_TRUE;
}

static void *spInitPluginInstanceMonkey(const char *lang)
{
    spMonkeyPluginInstance instance;
    
    instance = xalloc(1, struct _spMonkeyPluginInstance);

    instance->pAPEDecompress = NULL;
    instance->pAPECompress = NULL;
    
    instance->compression_level = APE_COMPRESSION_LEVEL_HIGH / 1000;

    instance->num_channel = 2;
    instance->samp_rate = 44100;
    instance->samp_bit = 16;
    
    instance->current_pos = 0;
    instance->total_length = 0;

    spInitSongInfo(&instance->song_info);

    instance->buffer_size = MONKEY_DEFAULT_BUFFER_SIZE;
    instance->buffer = xalloc(instance->buffer_size, unsigned char);
    instance->buffer_samp_byte = (instance->samp_bit / 8);
    instance->buffer_length = instance->buffer_size / instance->buffer_samp_byte;
    
    return (void *)instance;
}

static spBool spFreePluginInstanceMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;

    xfree(pinstance->buffer);
    xfree(pinstance);
    
    return SP_TRUE;
}

static spBool spSetPluginFileTypeMonkey(void *instance, int index)
{
    return SP_TRUE;
}

static int spGetPluginFileTypeMonkey(void *instance)
{
    return 0;
}

static spBool spSetPluginSongInfoMonkey(void *instance, spSongInfo *song_info)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    spCopySongInfo(&pinstance->song_info, song_info);

    return SP_TRUE;
}

static spBool spGetPluginSongInfoMonkey(void *instance, spSongInfo *song_info)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;

    spCopySongInfo(song_info, &pinstance->song_info);

    return SP_TRUE;
}

static const char *spGetPluginBestSuffixMonkey(void *instance)
{
    return ".ape";
}

static spBool spSetPluginSampleBitMonkey(void *instance, int samp_bit)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;

    if (samp_bit != 16/* && samp_bit != 8*/&& samp_bit != 24 && samp_bit != 32
#if APE_VERSION_MAJOR >= 10
        && samp_bit != 33
#endif
        ) {
	return SP_FALSE;
    }
    pinstance->samp_bit = samp_bit;
    pinstance->buffer_samp_byte = (pinstance->samp_bit / 8);
    pinstance->buffer_length = pinstance->buffer_size / pinstance->buffer_samp_byte;

    return SP_TRUE;
}

static spBool spGetPluginSampleBitMonkey(void *instance, int *samp_bit)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;

    *samp_bit = pinstance->samp_bit;

    return SP_TRUE;
}

static spBool spSetPluginChannelMonkey(void *instance, int num_channel)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;

    pinstance->num_channel = num_channel;
    
    return SP_TRUE;
}

static spBool spGetPluginChannelMonkey(void *instance, int *num_channel)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    *num_channel = pinstance->num_channel;

    return SP_TRUE;
}

static spBool spSetPluginSampleRateMonkey(void *instance, double samp_rate)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;

    pinstance->samp_rate = (long)spRound(samp_rate);

    return SP_TRUE;
}

static spBool spGetPluginSampleRateMonkey(void *instance, double *samp_rate)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    *samp_rate = (double)pinstance->samp_rate;

    return SP_TRUE;
}

static spBool spSetPluginOtherInfoMonkey(void *instance, const char *id, char *data)
{
    
    return SP_FALSE;
}

static spBool spGetPluginOtherInfoMonkey(void *instance, const char *id, char *data)
{

    return SP_FALSE;
}

static int convertStringToAPEString(const char *string, APEChar *apestring, int size)
{
#if APE_FILE_VERSION_NUMBER >= 3980
    apestring[size - 1] = 0;
    return mbstowcs((wchar_t *)apestring, string, size - 1);
#else
    spStrCopy(apestring, size, string);
    return strlen(apestring);
#endif
}
